import attr
import collections
import trio
from trio_typing import TaskStatus
from functools import partial
from async_generator import asynccontextmanager
from typing import Any, AsyncIterator, Awaitable, Callable, Optional, overload
from ._multi_cancel import MultiCancelScope


def _get_coroutine_or_flag_problem(
    async_fn: Callable[..., Awaitable[Any]], *args: Any, **kwargs: Any
) -> Awaitable[Any]:
    """Call async_fn(*args) to produce and return a coroutine. If that
    doesn't work or doesn't produce a coroutine, try to get help
    from trio in describing what went wrong.
    """
    try:
        # can we call it?
        coro = async_fn(*args, **kwargs)
    except TypeError:
        probe_fn = async_fn
    else:
        # did we get a coroutine object back?
        if isinstance(coro, collections.abc.Coroutine):
            return coro
        probe_fn = partial(async_fn, **kwargs)

    # TODO: upstream a change that lets us access just the nice
    # error detection logic without running the risk of starting a task

    # If we're not happy with this async_fn, trio won't be either,
    # and will tell us why in much greater detail.
    try:
        trio.lowlevel.spawn_system_task(probe_fn, *args)
    except TypeError as ex:
        problem_with_async_fn = ex
    else:
        # we started the task successfully, wtf?
        raise trio.TrioInternalError(
            "tried to spawn a dummy task to figure out what was wrong with "
            "{async_fn!r} as an async function, but it seems to have started "
            "successfully -- all bets are off at this point"
        )
    raise problem_with_async_fn


@asynccontextmanager
async def open_service_nursery() -> AsyncIterator[trio.Nursery]:
    """Provides a nursery augmented with a cancellation ordering constraint.

    If an entire service nursery becomes cancelled, either due to an
    exception raised by some task in the nursery or due to the
    cancellation of a scope that surrounds the nursery, the body of
    the nursery ``async with`` block will receive the cancellation
    first, and no other tasks in the nursery will be cancelled until
    the body of the ``async with`` block has been exited.

    This is intended to support the common pattern where the body of
    the ``async with`` block uses some service that the other
    task(s) in the nursery provide. For example, if you have::

        async with open_websocket(host, port) as conn:
            await communicate_with_websocket(conn)

    where ``open_websocket()`` enters a nursery and spawns some tasks
    into that nursery to manage the connection, you probably want
    ``conn`` to remain usable in any ``finally`` or ``__aexit__``
    blocks in ``communicate_with_websocket()``.  With a regular
    nursery, this is not guaranteed; with a service nursery, it is.
    An example hinting at general usage::

        @asynccontextmanager
        async def open_websocket(host, port):
            async with open_service_nursery() as nursery:
                try:
                    # ... make some child tasks ...
                    yield connection
                finally:
                    # The yield body is already cancelled, and
                    # child tasks are still available here for cleanup...
                    pass

    Now, anything in the body of the ``open_websocket()`` context, including
    ``communicate_with_websocket()``, will be given first opportunity to cancel
    gracefully.  Subsequently, the ``finally`` block in the ``open_websocket()``
    implementation runs, and tasks spawned within the ``try`` body are still
    available during cleanup.

    Note that child tasks spawned using ``start()`` gain their protection from
    premature cancellation only at the point of their call to
    ``task_status.started()``.
    """

    async with trio.open_nursery() as nursery:
        child_task_scopes = MultiCancelScope(shield=True)

        def start_soon(
            async_fn: Callable[..., Awaitable[Any]],
            *args: Any,
            name: Optional[str] = None,
        ) -> None:
            async def wrap_child(coro: Awaitable[Any]) -> None:
                with child_task_scopes.open_child():
                    await coro

            coro = _get_coroutine_or_flag_problem(async_fn, *args)
            type(nursery).start_soon(nursery, wrap_child, coro, name=name or async_fn)

        async def start(
            async_fn: Callable[..., Awaitable[Any]],
            *args: Any,
            name: Optional[str] = None,
        ) -> Any:
            async def wrap_child(*, task_status: TaskStatus[Any]) -> None:
                # For start(), the child doesn't get shielded until it
                # calls task_status.started().
                shield_scope = child_task_scopes.open_child(shield=False)

                def wrap_started(value: object = None) -> None:
                    type(task_status).started(task_status, value)  # type: ignore
                    if trio.lowlevel.current_task().parent_nursery is not nursery:
                        # started() didn't move the task due to a cancellation,
                        # so it doesn't get the shield
                        return
                    shield_scope.shield = child_task_scopes.shield

                task_status.started = wrap_started  # type: ignore
                with shield_scope:
                    await async_fn(*args, task_status=task_status)

            return await type(nursery).start(nursery, wrap_child, name=name or async_fn)

        nursery.start_soon = start_soon  # type: ignore
        nursery.start = start  # type: ignore
        try:
            yield nursery
        finally:
            child_task_scopes.shield = False



import attr
import trio
from async_generator import asynccontextmanager
from collections import OrderedDict
from typing import (
    AsyncIterator,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Set,
    TYPE_CHECKING,
)


@attr.s(auto_attribs=True)
class _RWLockStatistics:
    locked: str
    readers: FrozenSet[trio.lowlevel.Task]
    writer: Optional[trio.lowlevel.Task]
    readers_waiting: int
    writers_waiting: int


@attr.s(eq=False, repr=False)
class RWLock:
    """A `readers-writer lock
    <https://en.wikipedia.org/wiki/Readers%E2%80%93writer_lock>`__.

    Each acquisition of the lock specifies whether it is a "reader" or
    a "writer". At any given time, the lock may be held by one writer
    and no readers, by many readers and no writer, or by no one.

    This implementation is fair by default: if task A tried to acquire
    the lock before task B did, task B won't get it first. This
    implies that new readers can't acquire a reader-held lock after a
    writer has started waiting to acquire it, which helps avoid
    starvation of writers by readers. (The Wikipedia article linked
    above calls this "write-preferring".) If you want different behavior,
    see the :attr:`read_biased` attribute.

    Attributes:
      read_biased (bool): Whether new readers should be able to
          immediately acquire a readers-held lock even after some
          writers have started waiting for it. (The Wikipedia article
          linked above calls this "weakly read-preferring".) Note that
          setting :attr:`read_biased` to :data:`True` can result in
          indefinite starvation of writers if the read workload is
          busy enough.  Changing this attribute to :data:`True` will
          immediately wake up all waiting readers to grant them the
          lock if it is currently readers-held with writers waiting.

    """

    _writer: Optional[trio.lowlevel.Task] = attr.ib(default=None, init=False)
    _readers: Set[trio.lowlevel.Task] = attr.ib(factory=set, init=False)
    _waiting: "OrderedDict[trio.lowlevel.Task, bool]" = attr.ib(
        factory=OrderedDict, init=False
    )
    _waiting_writers_count: int = attr.ib(default=0, init=False)
    _read_biased: bool = attr.ib(default=False, kw_only=True)

    def __repr__(self) -> str:
        state = (
            "write-locked"
            if self._writer
            else "read-locked"
            if self._readers
            else "unlocked"
        )
        if self._read_biased:
            state += " read-biased"
        if self._waiting:
            waiters_descs: List[str] = []
            if self._waiting_writers_count:
                waiters_descs.append(f"{self._waiting_writers_count} writers")
            readers_count = len(self._waiting) - self._waiting_writers_count
            if readers_count:
                waiters_descs.append(f"{readers_count} readers")
            waiters = ", {} waiting".format(" and ".join(waiters_descs))
        else:
            waiters = ""
        return f"<{state} RWLock object at {id(self):#x}{waiters}>"

    def locked(self) -> str:
        """Check whether the lock is currently held.

        Returns:
          ``"read"`` if the lock is held by reader(s), ``"write"``
          if the lock is held by a writer, or ``""`` (which tests
          as false) if the lock is not held.
        """
        return "read" if self._readers else "write" if self._writer else ""

    @trio.lowlevel.enable_ki_protection
    def acquire_nowait(self, *, for_write: bool) -> None:
        """Attempt to acquire the lock, without blocking.

        Args:
          for_write: If True, attempt to acquire the lock in write mode,
              which provides exclusive access. If False, attempt to acquire the
              lock in read mode, which permits other readers to also hold it.

        Raises:
          trio.WouldBlock: if the lock cannot be acquired without blocking
          RuntimeError: if the current task already holds the lock (in either
              read or write mode)
        """
        task = trio.lowlevel.current_task()
        if self._writer is task or task in self._readers:
            raise RuntimeError("attempt to re-acquire an already held RWLock")
        if self._writer is not None:
            raise trio.WouldBlock

        if for_write and not self._readers:
            self._writer = task
        elif not for_write and (self._read_biased or not self._waiting_writers_count):
            self._readers.add(task)
        else:
            raise trio.WouldBlock

    @trio.lowlevel.enable_ki_protection
    async def acquire(self, *, for_write: bool) -> None:
        """Acquire the lock, blocking if necessary.

        Args:
          for_write: If True, acquire the lock in write mode,
              which provides exclusive access. If False, acquire the
              lock in read mode, which permits other readers to also hold it.

        Raises:
          RuntimeError: if the current task already holds the lock (in either
              read or write mode)
        """
        await trio.lowlevel.checkpoint_if_cancelled()
        try:
            self.acquire_nowait(for_write=for_write)
        except trio.WouldBlock:
            task = trio.lowlevel.current_task()
            self._waiting[task] = for_write
            self._waiting_writers_count += for_write

            def abort_fn(_: object) -> trio.lowlevel.Abort:
                del self._waiting[task]
                self._waiting_writers_count -= for_write
                return trio.lowlevel.Abort.SUCCEEDED

            await trio.lowlevel.wait_task_rescheduled(abort_fn)
        else:
            await trio.lowlevel.cancel_shielded_checkpoint()

    @trio.lowlevel.enable_ki_protection
    def release(self) -> None:
        """Release the lock.

        Raises:
          RuntimeError: if the current task does not hold the lock (in either
              read or write mode)
        """
        task = trio.lowlevel.current_task()
        if task is self._writer:
            self._writer = None
        elif task in self._readers:
            self._readers.remove(task)
            if self._readers:
                return
        else:
            raise RuntimeError("can't release a RWLock you don't own")

        while self._writer is None and self._waiting:
            task, for_write = self._waiting.popitem(last=False)
            if not for_write:
                # Next task is a reader: since we haven't woken
                # a writer yet, we can wake it,
                self._readers.add(task)
                trio.lowlevel.reschedule(task)
                # In read-biased mode we can continue to wake
                # all other readers.
                if self._read_biased:
                    self._wake_all_readers()
                    return
                # In fair mode we can only wake the readers that
                # arrived before the next writer, so keep iterating
                # through self._waiting.
            elif not self._readers:
                # Next task is a writer and there are no readers;
                # wake the writer and we're done.
                self._writer = task
                self._waiting_writers_count -= 1
                trio.lowlevel.reschedule(task)
                break
            else:
                # Next task is a writer, but can't be woken because
                # there are readers. Put it back at the front of the
                # line.
                self._waiting[task] = for_write
                self._waiting.move_to_end(task, last=False)
                break

    def _wake_all_readers(self) -> None:
        for task, for_write in list(self._waiting.items()):
            if not for_write:
                del self._waiting[task]
                self._readers.add(task)
                trio.lowlevel.reschedule(task)

    # https://github.com/python/mypy/issues/1362: mypy doesn't support
    # decorated properties yet
    if TYPE_CHECKING:
        read_biased: bool
    else:

        @property
        def read_biased(self) -> bool:
            return self._read_biased

        @read_biased.setter
        @trio.lowlevel.enable_ki_protection
        def read_biased(self, new_value: bool) -> None:
            if new_value and not self._read_biased:
                self._wake_all_readers()
            self._read_biased = new_value

    def acquire_read_nowait(self) -> None:
        """Equivalent to ``acquire_nowait(for_write=False)``."""
        return self.acquire_nowait(for_write=False)

    def acquire_write_nowait(self) -> None:
        """Equivalent to ``acquire_nowait(for_write=True)``."""
        return self.acquire_nowait(for_write=True)

    async def acquire_read(self) -> None:
        """Equivalent to ``acquire(for_write=False)``."""
        return await self.acquire(for_write=False)

    async def acquire_write(self) -> None:
        """Equivalent to ``acquire(for_write=True)``."""
        return await self.acquire(for_write=True)

    @trio.lowlevel.enable_ki_protection
    @asynccontextmanager
    async def read_locked(self) -> AsyncIterator[None]:
        """Returns an async context manager whose ``__aenter__`` blocks
        to acquire the lock in read mode, and whose ``__aexit__``
        synchronously releases it.
        """
        await self.acquire(for_write=False)
        try:
            yield
        finally:
            self.release()

    @trio.lowlevel.enable_ki_protection
    @asynccontextmanager
    async def write_locked(self) -> AsyncIterator[None]:
        """Returns an async context manager whose ``__aenter__`` blocks
        to acquire the lock in write mode, and whose ``__aexit__``
        synchronously releases it.
        """
        await self.acquire(for_write=True)
        try:
            yield
        finally:
            self.release()

    def statistics(self) -> _RWLockStatistics:
        r"""Return an object containing debugging information.

        Currently the following fields are defined:

        * ``locked``: boolean indicating whether the lock is held by anyone
        * ``state``: string with one of the values ``"read"`` (held by one
          or more readers), ``"write"`` (held by one writer),
          or ``"unlocked"`` (held by no one)
        * ``readers``: a frozenset of the :class:`~trio.lowlevel.Task`\s
          currently holding the lock in read mode (may be empty)
        * ``writer``: the :class:`trio.lowlevel.Task` currently holding
          the lock in write mode, or None if the lock is not held in write mode
        * ``readers_waiting``: the number of tasks blocked waiting to acquire
          the lock in read mode
        * ``writers_waiting``: the number of tasks blocked waiting to acquire
          the lock in write mode

        """
        return _RWLockStatistics(
            locked=self.locked(),
            readers=frozenset(self._readers),
            writer=self._writer,
            readers_waiting=len(self._waiting) - self._waiting_writers_count,
            writers_waiting=self._waiting_writers_count,
        )



from ._version import __version__

from ._rwlock import RWLock as RWLock
from ._streams import (
    BufferedReceiveStream as BufferedReceiveStream,
    TextReceiveStream as TextReceiveStream,
)
from ._multi_cancel import MultiCancelScope as MultiCancelScope
from ._service_nursery import open_service_nursery as open_service_nursery
from ._meta import ScopedObject as ScopedObject, BackgroundObject as BackgroundObject

# watch this space...

_export = None
for _export in globals().values():
    if hasattr(_export, "__module__"):
        _export.__module__ = __name__
del _export



import pytest
from typing import Any
from trio_typing import TaskStatus

import trio
import trio.testing
from .. import open_service_nursery


async def test_basic(autojump_clock: trio.testing.MockClock) -> None:
    record = []
    async with open_service_nursery() as nursery:

        @nursery.start_soon
        async def background_task() -> None:
            try:
                await trio.sleep_forever()
            finally:
                record.append("background_task exiting")

        (task,) = nursery.child_tasks
        assert "background_task" in task.name

        nursery.cancel_scope.cancel()
        with trio.CancelScope(shield=True):
            await trio.sleep(1)
        record.append("body exiting")
        await trio.sleep(0)
        pytest.fail("should've been cancelled")  # pragma: no cover

    assert nursery.cancel_scope.cancelled_caught
    assert record == ["body exiting", "background_task exiting"]


async def test_start(autojump_clock: trio.testing.MockClock) -> None:
    record = []

    async def sleep_then_start(val: int, *, task_status: TaskStatus[int]) -> None:
        await trio.sleep(1)
        task_status.started(val)
        try:
            await trio.sleep(10)
            record.append("background task finished")  # pragma: no cover
        finally:
            record.append("background task exiting")

    async def shielded_sleep_then_start(*, task_status: TaskStatus[None]) -> None:
        with trio.CancelScope(shield=True):
            await trio.sleep(1)
        task_status.started()
        await trio.sleep(10)

    async with open_service_nursery() as nursery:
        # Child can be cancelled normally while it's starting
        with trio.move_on_after(0.5) as scope:
            await nursery.start(sleep_then_start, 1)
        assert scope.cancelled_caught
        assert not nursery.child_tasks

        # If started() is the first thing to notice a cancellation, the task
        # stays in the old nursery and remains unshielded
        with trio.move_on_after(0.5) as scope:
            await nursery.start(shielded_sleep_then_start)
        assert scope.cancelled_caught
        assert not nursery.child_tasks

        assert trio.current_time() == 1.5

        # Otherwise, once started() is called the child is shielded until
        # the 'async with' block exits.
        assert 42 == await nursery.start(sleep_then_start, 42)
        assert trio.current_time() == 2.5

        nursery.cancel_scope.cancel()
        with trio.CancelScope(shield=True):
            await trio.sleep(1)
        record.append("parent task finished")

    assert trio.current_time() == 3.5
    assert record == ["parent task finished", "background task exiting"]


async def test_problems() -> None:
    async with open_service_nursery() as nursery:
        with pytest.raises(TypeError) as info:
            nursery.start_soon(trio.sleep)
        assert "missing 1 required positional argument" in str(info.value)

        with pytest.raises(TypeError) as info:
            nursery.start_soon(trio.sleep(1))  # type: ignore
        assert "Trio was expecting an async function" in str(info.value)

        with pytest.raises(TypeError) as info:
            nursery.start_soon(int, 42)  # type: ignore
        assert "appears to be synchronous" in str(info.value)

        first_call = True

        def evil() -> Any:
            nonlocal first_call
            if first_call:
                first_call = False
                return 42
            else:
                return trio.sleep(0)

        with pytest.raises(trio.TrioInternalError) as info2:
            nursery.start_soon(evil)
        assert "all bets are off at this point" in str(info2.value)



import attr
import pytest
import types
import trio
import trio.testing
from async_generator import asynccontextmanager
from typing import Any, AsyncIterator, Coroutine, Iterator, List
from trio_typing import TaskStatus

from .. import ScopedObject, BackgroundObject


def test_too_much_magic() -> None:
    with pytest.raises(TypeError) as info:

        class TooMuchMagic(ScopedObject):  # pragma: no cover
            async def __open__(self) -> None:
                pass

            @asynccontextmanager
            async def __wrap__(self) -> AsyncIterator[None]:
                yield

    assert str(info.value) == (
        "ScopedObjects can define __open__/__close__, or __wrap__, but not both"
    )


@types.coroutine
def async_yield(value: str) -> Iterator[str]:
    yield value


def test_mro() -> None:
    class A(ScopedObject):
        async def __open__(self) -> None:
            await async_yield("open A")

    class B(A):
        async def __open__(self) -> None:
            await async_yield("open B")

        async def __close__(self) -> None:
            await async_yield("close B")

    class C(A):
        async def __open__(self) -> None:
            await async_yield("open C")

        async def __close__(self) -> None:
            await async_yield("close C")

    class D(B, C):
        def __init__(self, value: int):
            self.value = value

        async def __close__(self) -> None:
            await async_yield("close D")

    assert D.__mro__ == (D, B, C, A, ScopedObject, object)
    d_mgr = D(42)
    assert not isinstance(d_mgr, D)
    assert not hasattr(d_mgr, "value")
    assert hasattr(d_mgr, "__aenter__")

    async def use_it() -> None:
        async with d_mgr as d:
            assert isinstance(d, D)
            assert d.value == 42
            await async_yield("body")

    coro: Coroutine[str, None, None] = use_it()
    record = []
    while True:
        try:
            record.append(coro.send(None))
        except StopIteration:
            break
    assert record == [
        "open A",
        "open C",
        "open B",
        "body",
        "close D",
        "close B",
        "close C",
    ]


@attr.s(auto_attribs=True)
class Example(BackgroundObject):
    ticks: int = 0
    record: List[str] = attr.Factory(list)
    exiting: bool = False

    def __attrs_post_init__(self) -> None:
        assert not hasattr(self, "nursery")
        self.record.append("attrs_post_init")

    async def __open__(self) -> None:
        self.record.append("open")
        await self.nursery.start(self._background_task)
        self.record.append("started")

    async def __close__(self) -> None:
        assert len(self.nursery.child_tasks) != 0
        # Make sure this doesn't raise AttributeError in aexit:
        del self.nursery
        self.record.append("close")
        self.exiting = True

    async def _background_task(self, *, task_status: TaskStatus[None]) -> None:
        self.record.append("background")
        await trio.sleep(1)
        self.record.append("starting")
        task_status.started()
        self.record.append("running")
        while not self.exiting:
            await trio.sleep(1)
            self.ticks += 1
        self.record.append("stopping")


class DaemonExample(Example, daemon=True):
    pass


async def test_background(autojump_clock: trio.testing.MockClock) -> None:
    async with Example(ticks=100) as obj:
        assert obj.record == [
            "attrs_post_init",
            "open",
            "background",
            "starting",
            "running",
            "started",
        ]
        del obj.record[:]
        await trio.sleep(5.5)
    assert obj.record == ["close", "stopping"]
    # 1 sec start + 6 ticks
    assert trio.current_time() == 7.0
    assert obj.ticks == 106
    assert not hasattr(obj, "nursery")

    # With daemon=True, the background tasks are cancelled when the parent exits
    async with DaemonExample() as obj2:
        assert obj2.record == [
            "attrs_post_init",
            "open",
            "background",
            "starting",
            "running",
            "started",
        ]
        del obj2.record[:]
        await trio.sleep(5.5)
    assert obj2.record == ["close"]
    assert trio.current_time() == 13.5
    assert obj2.ticks == 5



import pytest
import itertools
import trio
import trio.testing
from .. import RWLock
from typing import List, Optional
from trio_typing import TaskStatus


async def test_rwlock(autojump_clock: trio.testing.MockClock) -> None:
    lock = RWLock()
    assert not lock.locked()

    lock.acquire_read_nowait()
    assert lock.locked() == "read"

    with pytest.raises(RuntimeError):
        lock.acquire_read_nowait()
    with pytest.raises(RuntimeError):
        lock.acquire_write_nowait()
    lock.release()
    with pytest.raises(RuntimeError):
        lock.release()

    with trio.testing.assert_checkpoints():
        await lock.acquire_write()
    assert lock.locked() == "write"
    with pytest.raises(RuntimeError):
        await lock.acquire_read()
    with pytest.raises(RuntimeError):
        await lock.acquire_write()
    lock.release()

    async with lock.read_locked():
        assert lock.locked() == "read"

    async with lock.write_locked():
        assert lock.locked() == "write"

    start_order = itertools.count()
    acquire_times: List[Optional[float]] = [None] * 10

    async def holder_task(
        for_write: bool, task_status: TaskStatus[trio.lowlevel.Task]
    ) -> None:
        my_slot = next(start_order)
        repr(lock)  # smoke test
        task_status.started(trio.lowlevel.current_task())
        await lock.acquire(for_write=for_write)
        acquire_times[my_slot] = trio.current_time()
        try:
            await trio.sleep(1)
        finally:
            lock.release()

    async with trio.open_nursery() as nursery:
        t0 = await nursery.start(holder_task, True)
        t1a = await nursery.start(holder_task, False)
        t1b = await nursery.start(holder_task, False)
        t1c = await nursery.start(holder_task, False)
        await nursery.start(holder_task, True)  # t2
        await nursery.start(holder_task, False)  # t3a
        await nursery.start(holder_task, False)  # t3b
        await nursery.start(holder_task, True)  # t4
        await nursery.start(holder_task, True)  # t5
        t6 = await nursery.start(holder_task, False)

        await trio.sleep(0.5)
        assert "write-locked" in repr(lock)
        assert lock.statistics().__dict__ == {
            "locked": "write",
            "readers": frozenset(),
            "writer": t0,
            "readers_waiting": 6,
            "writers_waiting": 3,
        }
        with pytest.raises(RuntimeError):
            lock.release()
        with pytest.raises(trio.WouldBlock):
            lock.acquire_read_nowait()
        with pytest.raises(trio.WouldBlock):
            lock.acquire_write_nowait()

        await trio.sleep(1)
        assert "read-locked" in repr(lock)
        assert lock.statistics().__dict__ == {
            "locked": "read",
            "readers": frozenset([t1a, t1b, t1c]),
            "writer": None,
            "readers_waiting": 3,
            "writers_waiting": 3,
        }
        with pytest.raises(RuntimeError):
            lock.release()
        with pytest.raises(trio.WouldBlock):
            # even in read state, can't acquire for read if writers are waiting
            lock.acquire_read_nowait()
        with pytest.raises(trio.WouldBlock):
            lock.acquire_write_nowait()

        await trio.sleep(5)
        assert "read-locked" in repr(lock)
        assert lock.statistics().__dict__ == {
            "locked": "read",
            "readers": frozenset([t6]),
            "writer": None,
            "readers_waiting": 0,
            "writers_waiting": 0,
        }
        lock.acquire_read_nowait()
        lock.release()
        with pytest.raises(trio.WouldBlock):
            lock.acquire_write_nowait()

    assert acquire_times == pytest.approx([0, 1, 1, 1, 2, 3, 3, 4, 5, 6])

    # test cancellation
    start_order = itertools.count()
    async with trio.open_nursery() as nursery:
        await nursery.start(holder_task, True)
        await nursery.start(holder_task, True)
        await nursery.start(holder_task, False)
        await nursery.start(holder_task, False)
        await nursery.start(holder_task, False)
        await nursery.start(holder_task, True)
        await nursery.start(holder_task, False)
        await nursery.start(holder_task, False)
        await nursery.start(holder_task, True)
        await nursery.start(holder_task, True)
        await nursery.start(holder_task, False)

        await trio.sleep(0.5)
        nursery.cancel_scope.cancel()

    assert nursery.cancel_scope.cancelled_caught
    assert trio.current_time() == pytest.approx(7.5)
    assert "unlocked" in repr(lock)
    assert lock.statistics().__dict__ == {
        "locked": "",
        "readers": frozenset(),
        "writer": None,
        "readers_waiting": 0,
        "writers_waiting": 0,
    }


async def test_read_biased(autojump_clock: trio.testing.MockClock) -> None:
    lock = RWLock(read_biased=True)
    assert "read-biased" in repr(lock)
    assert lock.read_biased

    async def holder_task(
        for_write: bool, task_status: TaskStatus[trio.lowlevel.Task]
    ) -> None:
        task_status.started(trio.lowlevel.current_task())
        await lock.acquire(for_write=for_write)
        try:
            await trio.sleep(1)
        finally:
            lock.release()

    async with trio.open_nursery() as nursery:
        t1a = await nursery.start(holder_task, False)
        t1b = await nursery.start(holder_task, False)
        t2 = await nursery.start(holder_task, True)
        t1c = await nursery.start(holder_task, False)

        # Reader (t1c) that arrives after the writer (t2) can get the
        # lock immediately, before the writer does
        await trio.sleep(0.5)
        assert lock.statistics().readers == frozenset([t1a, t1b, t1c])
        assert lock.statistics().writer is None

        await trio.sleep(1)
        assert lock.statistics().readers == frozenset()
        assert lock.statistics().writer is t2

        # If an additional writer gets in line for a writer-held lock
        # before any readers do, they get it before readers that come later
        # (i.e., the bias towards readers is "weak" -- "strong" would wake
        # up all readers when any writer released).
        t3 = await nursery.start(holder_task, True)
        t4a = await nursery.start(holder_task, False)
        t5 = await nursery.start(holder_task, True)
        t4b = await nursery.start(holder_task, False)

        await trio.sleep(1)
        assert lock.statistics().readers == frozenset()
        assert lock.statistics().writer is t3

        await trio.sleep(1)
        assert lock.statistics().readers == frozenset([t4a, t4b])
        assert lock.statistics().writer is None

        await trio.sleep(1)
        assert lock.statistics().readers == frozenset()
        assert lock.statistics().writer is t5

        # Read-bias can be turned on and off dynamically
        lock.read_biased = False
        assert not lock.read_biased
        t6 = await nursery.start(holder_task, False)
        t7 = await nursery.start(holder_task, True)
        t8 = await nursery.start(holder_task, False)

        # Now reader t8 has to wait because writer t7 arrived first,
        # even though the lock is held by reader t6
        await trio.sleep(1)
        assert lock.statistics().readers == frozenset([t6])
        assert lock.statistics().writer is None

        # If we turn on read-bias again, t8 immediately gets the lock
        lock.read_biased = True
        assert lock.statistics().readers == frozenset([t6, t8])
        assert lock.statistics().writer is None

        await trio.sleep(0.75)
        assert lock.statistics().readers == frozenset([t8])
        assert lock.statistics().writer is None

        await trio.sleep(0.75)
        assert lock.statistics().readers == frozenset()
        assert lock.statistics().writer is t7



import pytest

import trio
import trio.testing
from .. import MultiCancelScope


async def test_basic(autojump_clock: trio.testing.MockClock) -> None:
    parent = MultiCancelScope()
    finish_order = []

    async def cancel_child_before_entering() -> None:
        child = parent.open_child()
        assert not child.cancel_called
        child.cancel()
        assert child.cancel_called
        assert not child.cancelled_caught
        await trio.sleep(0.2)
        with child:
            assert not child.cancelled_caught
            await trio.sleep(1)
        assert child.cancelled_caught
        finish_order.append("cancel_child_before_entering")

    async def cancel_child_after_entering() -> None:
        with parent.open_child() as child:
            await trio.sleep(0.3)
            child.cancel()
            await trio.sleep(1)
        assert child.cancel_called
        assert child.cancelled_caught
        finish_order.append("cancel_child_after_entering")

    async def cancel_child_via_local_deadline() -> None:
        child = parent.open_child()
        child.deadline = trio.current_time() + 0.4
        deadline_before_entering = child.deadline
        with child:
            assert child.deadline == deadline_before_entering
            await trio.sleep(1)
        assert child.cancel_called
        assert child.cancelled_caught
        finish_order.append("cancel_child_via_local_deadline")

    async def cancel_child_via_local_deadline_2() -> None:
        child = parent.open_child()
        child.deadline = trio.current_time() + 1.0
        with child:
            child.deadline -= 0.9
            await trio.sleep(1)
        assert child.cancel_called
        assert child.cancelled_caught
        finish_order.append("cancel_child_via_local_deadline_2")

    async def cancel_parent_before_entering() -> None:
        child = parent.open_child()
        await trio.sleep(0.6)
        assert child.cancel_called
        assert not child.cancelled_caught
        with child:
            await trio.sleep(1)
        assert child.cancelled_caught
        finish_order.append("cancel_parent_before_entering")

    async def cancel_parent_after_entering() -> None:
        with parent.open_child() as child:
            await trio.sleep(1)
        assert child.cancel_called
        assert child.cancelled_caught
        finish_order.append("cancel_parent_after_entering")

    async with trio.open_nursery() as nursery:
        nursery.start_soon(cancel_child_before_entering)
        nursery.start_soon(cancel_child_after_entering)
        nursery.start_soon(cancel_child_via_local_deadline)
        nursery.start_soon(cancel_child_via_local_deadline_2)
        nursery.start_soon(cancel_parent_before_entering)
        nursery.start_soon(cancel_parent_after_entering)
        await trio.sleep(0.5)
        assert "MultiCancelScope cancelled" not in repr(parent)
        assert not parent.cancel_called
        parent.cancel()
        assert parent.cancel_called
        assert "MultiCancelScope cancelled" in repr(parent)
        parent.cancel()
        await trio.sleep(0.2)

        nursery.cancel_scope.deadline = trio.current_time() + 0.1
        with parent.open_child() as child:
            child.deadline = nursery.cancel_scope.deadline
            assert child.cancel_called
            assert not child.cancelled_caught
            await trio.sleep_forever()
        assert child.cancelled_caught
        finish_order.append("cancel_parent_before_creating")

    assert not nursery.cancel_scope.cancelled_caught
    assert finish_order == [
        "cancel_child_via_local_deadline_2",  # t+0.1
        "cancel_child_before_entering",  # t+0.2
        "cancel_child_after_entering",  # t+0.3
        "cancel_child_via_local_deadline",  # t+0.4
        "cancel_parent_after_entering",  # t+0.5
        "cancel_parent_before_entering",  # t+0.6
        "cancel_parent_before_creating",  # t+0.7
    ]


async def test_shielding(autojump_clock: trio.testing.MockClock) -> None:
    parent = MultiCancelScope()
    finish_order = []

    async def shield_child_on_creation() -> None:
        try:
            with parent.open_child(shield=True):
                await trio.sleep(1)
            assert False  # pragma: no cover
        finally:
            finish_order.append("shield_child_on_creation")

    async def shield_child_before_entering() -> None:
        child = parent.open_child()
        child.shield = True
        try:
            with child:
                await trio.sleep(1)
            assert False  # pragma: no cover
        finally:
            with trio.CancelScope(shield=True):
                await trio.sleep(0.1)
            finish_order.append("shield_child_before_entering")

    async def shield_child_after_entering() -> None:
        try:
            with parent.open_child() as child:
                child.shield = True
                await trio.sleep(1)
            assert False  # pragma: no cover
        finally:
            with trio.CancelScope(shield=True):
                await trio.sleep(0.2)
            finish_order.append("shield_child_after_entering")

    async def shield_child_when_parent_shielded() -> None:
        try:
            with trio.CancelScope(shield=True):
                await trio.sleep(0.3)
            with parent.open_child():
                await trio.sleep(1)
        finally:
            with trio.CancelScope(shield=True):
                await trio.sleep(0.3)
            finish_order.append("shield_child_when_parent_shielded")

    async def shield_child_after_parent_unshielded() -> None:
        with parent.open_child(shield=True) as child:
            this_task = trio.lowlevel.current_task()

            def abort_fn(_):  # type: ignore
                trio.lowlevel.reschedule(this_task)
                return trio.lowlevel.Abort.FAILED

            await trio.lowlevel.wait_task_rescheduled(abort_fn)
            child.shield = True
            await trio.sleep(0.5)
        assert not child.cancelled_caught
        finish_order.append("shield_child_after_parent_unshielded")

    async with trio.open_nursery() as nursery:
        nursery.start_soon(shield_child_on_creation)
        nursery.start_soon(shield_child_before_entering)
        nursery.start_soon(shield_child_after_entering)
        nursery.start_soon(shield_child_when_parent_shielded)
        nursery.start_soon(shield_child_after_parent_unshielded)

        nursery.cancel_scope.cancel()
        assert parent.shield == False
        with trio.CancelScope(shield=True):
            await trio.sleep(0.2)
        assert "MultiCancelScope shielded" not in repr(parent)
        parent.shield = True
        assert "MultiCancelScope shielded" in repr(parent)
        assert parent.shield == True
        with trio.CancelScope(shield=True):
            await trio.sleep(0.2)
        parent.shield = False

    assert finish_order == [
        "shield_child_on_creation",  # t+0.4
        "shield_child_before_entering",  # t+0.5
        "shield_child_after_entering",  # t+0.6
        "shield_child_when_parent_shielded",  # t+0.7
        "shield_child_after_parent_unshielded",  # t+0.8
    ]



import pytest
import locale
import random
import sys

import trio
import trio.testing
from async_generator import asynccontextmanager
from functools import partial
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    List,
    Optional,
    Sequence,
    Tuple,
    cast,
)
from typing_extensions import AsyncContextManager
from .. import BufferedReceiveStream, TextReceiveStream


async def test_buffered_receive(autojump_clock: trio.testing.MockClock) -> None:
    send_stream, receive_stream_raw = trio.testing.memory_stream_one_way_pair()
    receive_stream = BufferedReceiveStream(receive_stream_raw, chunk_size=8)

    orig_receive_some = receive_stream_raw.receive_some
    raw_receive_sizes = []

    async def hooked_receive_some(max_bytes: int) -> bytes:
        result = await orig_receive_some(max_bytes)
        raw_receive_sizes.append(len(result))
        return result

    receive_stream_raw.receive_some = hooked_receive_some  # type: ignore

    # Send a big block, receive a little at a time
    data = bytes(val for val in range(64))
    await send_stream.send_all(data)

    async def checked_receive(num_bytes: int, raw_reads: Sequence[int] = ()) -> bytes:
        try:
            with trio.testing.assert_checkpoints():
                return await receive_stream.receive(num_bytes)
        finally:
            assert list(raw_reads) == raw_receive_sizes
            raw_receive_sizes[:] = []

    chunks = [
        await checked_receive(4, [8]),
        await checked_receive(1),
        await checked_receive(3),  # go exactly to end of chunk
        await checked_receive(1, [8]),
        await checked_receive(9, [8]),  # receive across chunk boundary
        await checked_receive(18, [12]),  # >1 chunk beyond buffered amount
        await checked_receive(12, [12]),  # >1 chunk starting from boundary
        await checked_receive(14, [14]),  # 2 bytes left after this point
        await checked_receive(1, [2]),
        await checked_receive(1),  # everything consumed
    ]
    assert b"".join(chunks) == data

    receive_stream.unget(b"1234")
    assert await receive_stream.receive(2) == b"12"
    receive_stream.unget(b"012")
    assert await receive_stream.receive(4) == b"0123"
    receive_stream.unget(b"tt")
    assert await receive_stream.receive(3) == b"tt4"

    with pytest.raises(trio.TooSlowError), trio.fail_after(1):
        await checked_receive(1)
    await send_stream.send_all(b"xyz")
    with pytest.raises(trio.TooSlowError), trio.fail_after(1):
        await checked_receive(10, [3])
    await send_stream.send_all(b"abcdabcdabcd")
    with pytest.raises(trio.TooSlowError), trio.fail_after(1):
        await checked_receive(16, [12])
    assert b"xyzabcd" == await checked_receive(7)
    await send_stream.aclose()
    assert b"abcdabcd" == await checked_receive(32, [0])
    assert b"" == await checked_receive(32, [0])

    # now try a clean EOF
    send_stream, receive_stream_raw = trio.testing.memory_stream_one_way_pair()
    receive_stream = BufferedReceiveStream(receive_stream_raw, chunk_size=8)
    orig_receive_some = receive_stream_raw.receive_some
    receive_stream_raw.receive_some = hooked_receive_some  # type: ignore

    await send_stream.send_all(b"1234")
    assert b"12" == await checked_receive(2, [4])
    with pytest.raises(trio.TooSlowError), trio.fail_after(1):
        await checked_receive(3)
    await send_stream.aclose()
    assert b"34" == await checked_receive(3, [0])

    await receive_stream.aclose()


@pytest.fixture
async def receiver_factory() -> AsyncIterator[
    Callable[[], AsyncContextManager[BufferedReceiveStream]]
]:
    async def send_task(send_stream: trio.abc.SendStream) -> None:
        for val in b"0123456789":
            await send_stream.send_all(bytes([val]))
            await trio.sleep(1)
        await send_stream.aclose()

    @asynccontextmanager
    async def receiver() -> AsyncIterator[BufferedReceiveStream]:
        send_stream, receive_stream_raw = trio.testing.memory_stream_one_way_pair()
        async with trio.open_nursery() as nursery:
            nursery.start_soon(send_task, send_stream)
            try:
                yield BufferedReceiveStream(receive_stream_raw, chunk_size=8)
            finally:
                nursery.cancel_scope.cancel()

    yield receiver


async def test_buffered_receive_helpers(
    autojump_clock: trio.testing.MockClock,
    receiver_factory: Callable[[], AsyncContextManager[BufferedReceiveStream]],
) -> None:
    async with receiver_factory() as receive_stream:
        assert b"012345" == await receive_stream.receive_all_or_none(6)
        assert b"6789" == await receive_stream.receive_all_or_none(4)
        assert None is await receive_stream.receive_all_or_none(42)

    async with receiver_factory() as receive_stream:
        assert b"0" == await receive_stream.receive_all_or_none(1)
        assert b"123456789" == await receive_stream.receive_all_or_none(9)
        assert None is await receive_stream.receive_all_or_none(3)

    async with receiver_factory() as receive_stream:
        assert b"012345" == await receive_stream.receive_all_or_none(6)
        with pytest.raises(ValueError) as info:
            await receive_stream.receive_all_or_none(5)
        assert str(info.value) == (
            "unclean EOF (4 bytes after boundary, expected at least 5)"
        )
        assert b"6789" == await receive_stream.receive(4)
        assert None is await receive_stream.receive_all_or_none(4)
        assert b"" == await receive_stream.receive(4)

    async with receiver_factory() as receive_stream:
        assert b"012345" == await receive_stream.receive_all_or_none(6)
        with pytest.raises(ValueError) as info:
            await receive_stream.receive_all_or_none(5)
        assert str(info.value) == (
            "unclean EOF (4 bytes after boundary, expected at least 5)"
        )
        assert b"6789" == await receive_stream.receive(4)
        assert None is await receive_stream.receive_all_or_none(4)
        assert b"" == await receive_stream.receive(4)

    async with receiver_factory() as receive_stream:
        assert b"0" == await receive_stream.receive_exactly(1)
        assert b"1234567" == await receive_stream.receive_exactly(7)
        with pytest.raises(ValueError) as info:
            await receive_stream.receive_exactly(3)
        assert str(info.value) == "unclean EOF (read only 2/3 bytes)"
        assert b"89" == await receive_stream.receive_exactly(2)
        with pytest.raises(ValueError) as info:
            await receive_stream.receive_exactly(3)
        assert str(info.value) == "unclean EOF (read only 0/3 bytes)"


async def test_text_receive(autojump_clock: trio.testing.MockClock) -> None:
    test_input = (
        b"The quick brown fox jumps over the lazy dog.\r\n\n\r\r\n\n"
        b"Yup.\n"
        b"That \xf0\x9f\xa6\x8a is still jumping.\r"  # fox emoji
    )

    # Test that encoding=None uses the locale preferred encoding
    stream = TextReceiveStream(None)  # type: ignore
    assert stream.encoding == locale.getpreferredencoding(False)
    del stream

    newline: Optional[str]
    for newline in ("\r", "\n", "\r\n", "", "jump", None):
        output_str = test_input.decode("utf-8")
        if newline == "":
            output_lines = output_str.splitlines(True)
        elif newline is None:
            output_lines = (
                output_str.replace("\r\n", "\n").replace("\r", "\n").splitlines(True)
            )
        else:
            output_lines = [line + newline for line in output_str.split(newline)]
            if output_lines[-1] == newline:
                del output_lines[-1]
            else:
                output_lines[-1] = output_lines[-1][: -len(newline)]

        Streams = Tuple[trio.testing.MemorySendStream, TextReceiveStream]

        async def make_streams_with_hook(
            hook: Optional[
                Callable[
                    [trio.testing.MemorySendStream, trio.testing.MemoryReceiveStream],
                    Awaitable[None],
                ]
            ]
        ) -> Streams:
            send_stream, receive_stream_raw = trio.testing.memory_stream_one_way_pair()
            receive_stream = TextReceiveStream(
                receive_stream_raw, "UTF-8", chunk_size=8, newline=newline
            )
            if hook is not None:
                send_stream.send_all_hook = partial(
                    hook, send_stream, receive_stream_raw
                )
            assert receive_stream.encoding == "UTF-8"
            assert receive_stream.errors is None
            return send_stream, receive_stream

        async def make_streams_all_at_once() -> Streams:
            return await make_streams_with_hook(None)

        async def make_streams_one_byte_at_a_time() -> Streams:
            async def trickle(
                send: trio.testing.MemorySendStream,
                receive: trio.testing.MemoryReceiveStream,
            ) -> None:
                while trio.testing.memory_stream_pump(send, receive, max_bytes=1):
                    await trio.sleep(1)

            return await make_streams_with_hook(trickle)

        async def make_streams_random_small_units() -> Streams:
            async def trickle(
                send: trio.testing.MemorySendStream,
                receive: trio.testing.MemoryReceiveStream,
            ) -> None:
                while trio.testing.memory_stream_pump(
                    send, receive, max_bytes=random.randint(1, 16)
                ):
                    await trio.sleep(1)

            return await make_streams_with_hook(trickle)

        for strategy in (
            make_streams_all_at_once,
            make_streams_one_byte_at_a_time,
            make_streams_random_small_units,
            make_streams_random_small_units,
            make_streams_random_small_units,
            make_streams_random_small_units,
            make_streams_random_small_units,
            make_streams_random_small_units,
        ):
            send_stream, receive_stream = await strategy()
            received_lines: List[str] = []

            async def receive_in_background() -> None:
                async for line in receive_stream:
                    received_lines.append(line)

            async with trio.open_nursery() as nursery:
                nursery.start_soon(receive_in_background)
                await send_stream.send_all(test_input)
                await trio.sleep(2)
                length_before_eof = len(received_lines)
                await send_stream.aclose()

            # Universal newline support will wait for a potential \n
            # after the trailing \r and we don't get that line until EOF.
            assert len(received_lines) == length_before_eof + (newline != "\r")
            assert received_lines == output_lines
            # The .newlines property is broken on PyPy:
            # https://bitbucket.org/pypy/pypy/issues/3012
            if sys.implementation.name == "cpython":
                if not newline:
                    newlines_seen = cast(Tuple[str, ...], receive_stream.newlines)
                    assert set(newlines_seen) == {"\r", "\n", "\r\n"}
                else:
                    assert receive_stream.newlines is None


async def test_text_receive_fix_errors() -> None:
    test_input = b"The quick brown \xf0\x9f\xa6\x8a jumps over the lazy dog.\n"
    for chunk_size in range(1, len(test_input)):
        send_stream, receive_stream_raw = trio.testing.memory_stream_one_way_pair()
        receive_stream = TextReceiveStream(
            receive_stream_raw, chunk_size=chunk_size, encoding="ascii"
        )
        await send_stream.send_all(test_input)
        await send_stream.aclose()
        with pytest.raises(UnicodeDecodeError):
            await receive_stream.receive_line()
        with pytest.raises(UnicodeDecodeError):
            await receive_stream.receive_line()
        receive_stream.errors = "replace"
        assert receive_stream.errors == "replace"
        line = await receive_stream.receive_line()
        x = chr(65533)
        assert f"The quick brown {x}{x}{x}{x} jumps over the lazy dog.\n" == line
        assert "" == await receive_stream.receive_line()
        await receive_stream.aclose()


async def test_text_receive_hits_max_chars() -> None:
    test_input = b"The quick\nbrown \xf0\x9f\xa6\x8a\njumps over\r\nthe lazy dog.\n"
    for chunk_size in range(1, len(test_input)):
        send_stream, receive_stream_raw = trio.testing.memory_stream_one_way_pair()
        receive_stream = TextReceiveStream(
            receive_stream_raw, chunk_size=chunk_size, encoding="utf-8"
        )
        await send_stream.send_all(test_input)
        await send_stream.aclose()

        fox = chr(129_418)
        assert "The quick\n" == await receive_stream.receive_line(12)
        assert f"brown {fox}" == await receive_stream.receive_line(7)
        assert "\n" == await receive_stream.receive_line(10)
        assert "jumps ov" == await receive_stream.receive_line(8)
        assert "er\r" == await receive_stream.receive_line(3)
        assert "\n" == await receive_stream.receive_line(1)
        assert "the lazy dog.\n" == await receive_stream.receive_line(20)
        assert "" == await receive_stream.receive_line(1)



import abc
import functools
from async_generator import asynccontextmanager
from trio import Nursery
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    ClassVar,
    Dict,
    Optional,
    Type,
    TypeVar,
    TYPE_CHECKING,
)
from ._service_nursery import open_service_nursery


T = TypeVar("T")


class ScopedObjectMeta(abc.ABCMeta):
    # Metaclass that provides the ScopedObject magic. See ScopedObject
    # for the docs.
    def __new__(
        mcls, clsname: str, bases: Any, dct: Dict[str, Any], **kwargs: Any
    ) -> "ScopedObjectMeta":
        if "__open__" in dct or "__close__" in dct:
            if "__wrap__" in dct:
                raise TypeError(
                    "ScopedObjects can define __open__/__close__, or __wrap__, "
                    "but not both"
                )

            async def noop(self: Any) -> None:
                pass

            _open_: Callable[[Any], Awaitable[None]] = dct.get("__open__", noop)
            _close_: Callable[[Any], Awaitable[None]] = dct.get("__close__", noop)

            @asynccontextmanager
            async def wrap(self: Any) -> AsyncIterator[None]:
                async with super(cls, self).__wrap__():  # type: ignore
                    await _open_(self)
                    try:
                        yield
                    finally:
                        await _close_(self)

            wrap.__name__ = "__wrap__"
            wrap.__qualname__ = dct["__qualname__"] + ".__wrap__"
            dct["__wrap__"] = wrap

        # NB: wrap() closes over this 'cls' variable
        cls: ScopedObjectMeta = super().__new__(mcls, clsname, bases, dct, **kwargs)
        return cls

    @asynccontextmanager
    async def __call__(cls: Type[T], *args: Any, **kwds: Any) -> AsyncIterator[T]:
        self: T = super().__call__(*args, **kwds)  # type: ignore
        async with self.__wrap__():  # type: ignore
            yield self


class ScopedObject(metaclass=ScopedObjectMeta):
    """An object whose lifetime must be bound to an ``async with`` block.

    Suppose that ``Foo`` is a :class:`ScopedObject` subclass. Then if
    you say ``Foo(*args)``, you won't actually get a ``Foo`` object;
    instead, you'll get an async context manager that evaluates to a
    ``Foo`` object.  So you would need to say::

        async with Foo(*args) as my_foo:
            # do stuff with my_foo

    This allows ``Foo`` to have reliable control of its lifetime, so
    it can spawn background tasks, deterministically execute cleanup
    code, and so on.

    If you want to implement such an object, inherit from :class:`ScopedObject`
    and indicate what should happen on entry and exit of the context.
    This should be done in one of the following two ways:

    * Define async ``__open__`` and/or ``__close__`` methods, which will
      be called from the context ``__aenter__`` and ``__aexit__`` respectively,
      taking no arguments and returning ``None``.
      ``__close__`` will be called no matter whether the context exits
      normally or due to an exception. (It can tell whether there is an
      active exception by using :func:`sys.exc_info`, but cannot suppress
      it.) If you use this approach, :class:`ScopedObject` takes care of
      invoking any initialization and finalization logic
      supplied by your base classes.

    * Define a ``__wrap__`` method that returns an async context
      manager.  This gives you more flexibility than implementing
      ``__open__`` and ``__close__``, because you can run some code
      outside of your base classes' scope and can swallow exceptions,
      but means you have to enter the base classes' scope yourself.

    It is an error to define both ``__wrap__`` and (``__open__`` or
    ``__close__``). If you don't define ``__wrap__``,
    :class:`ScopedObject` generates it for you in terms of
    ``__open__`` and ``__close__``, with semantics equivalent to the
    following::

        @asynccontextmanager
        async def __wrap__(self):
            async with super().__wrap__():
                if hasattr(self, "__open__"):
                    await self.__open__()
                try:
                    yield
                finally:
                    if hasattr(self, "__close__"):
                        await self.__close__()

    """

    __slots__ = ("__weakref__",)

    @asynccontextmanager
    async def __wrap__(self) -> AsyncIterator[None]:
        yield

    if TYPE_CHECKING:
        # These are necessary to placate mypy, which doesn't understand
        # the asynccontextmanager metaclass __call__. They should never
        # actually get called.
        async def __aenter__(self: T) -> T:
            raise AssertionError

        async def __aexit__(self, *exc: object) -> None:
            raise AssertionError


class BackgroundObject(ScopedObject):
    """A :class:`ScopedObject` that automatically creates a
    :func:`service nursery <open_service_nursery>` for running background tasks.

    If you pass ``daemon=True`` when inheriting from :class:`BackgroundObject`,
    like so::

        class MyObject(BackgroundObject, daemon=True):
            ...

    then the tasks spawned in the nursery will automatically be cancelled
    when the ``async with MyObject(...) as obj:`` block exits.
    Otherwise, the parent waits for the children to exit normally, like
    the default Trio nursery behavior.

    """

    __slots__ = ("nursery",)
    __daemon: ClassVar[bool]

    def __init_subclass__(cls, *, daemon: bool = False, **kwargs: Any):
        cls.__daemon = daemon
        super().__init_subclass__(**kwargs)

    @asynccontextmanager
    async def __wrap__(self) -> AsyncIterator[None]:
        async with super().__wrap__():
            try:
                async with open_service_nursery() as nursery:
                    self.nursery = nursery
                    yield
                    if type(self).__daemon:
                        nursery.cancel_scope.cancel()
            finally:
                try:
                    del self.nursery
                except AttributeError:
                    pass



import attr
import codecs
import trio
from typing import Optional, Union, Tuple, AsyncIterator, TypeVar
from io import IncrementalNewlineDecoder


__all__ = ["BufferedReceiveStream", "TextReceiveStream"]


@attr.s(auto_attribs=True, eq=False)
class BufferedReceiveStream(trio.abc.AsyncResource):
    """Wraps a :class:`~trio.abc.ReceiveStream` with buffering capabilities,
    so you can receive known amounts of data at a time.
    """

    transport_stream: trio.abc.ReceiveStream
    chunk_size: int = 4096

    def __attrs_post_init__(self) -> None:
        self._buffer = bytearray()
        self._receive_pos = 0

    async def aclose(self) -> None:
        """Discard all buffered data and close the underlying stream."""
        del self._buffer[:]
        self._receive_pos = 0
        await self.transport_stream.aclose()

    async def receive(self, num_bytes: int) -> bytes:
        """Receive and return ``num_bytes`` bytes, or fewer if EOF is
        encountered.

        Args:
          num_bytes (int): The number of bytes to return. Must be
              greater than zero.

        Returns:
          bytes or bytearray: The data received, exactly ``num_bytes`` bytes
          unless EOF is encountered. If there is no data left to return
          before EOF, returns an empty bytestring (``b""``).

        Raises:
          Exception: Anything raised by the :meth:`~trio.abc.ReceiveStream.receive_some`
              method of the underlying transport stream.

        """
        if self._receive_pos + num_bytes > len(self._buffer):
            del self._buffer[: self._receive_pos]
            self._receive_pos = 0
            while num_bytes > len(self._buffer):
                to_receive = max(self.chunk_size, num_bytes - len(self._buffer))
                data = await self.transport_stream.receive_some(to_receive)
                if data == b"":
                    break  # EOF
                self._buffer.extend(data)
        else:
            await trio.lowlevel.checkpoint()

        data = self._buffer[self._receive_pos : self._receive_pos + num_bytes]
        self._receive_pos += len(data)
        return data

    async def receive_all_or_none(self, num_bytes: int) -> Optional[bytes]:
        """Receive and return exactly ``num_bytes`` bytes, or ``None``
        if EOF is encountered before receiving any bytes.

        Args:
          num_bytes (int): The number of bytes to return. Must be
              greater than zero.

        Returns:
          bytes or None:
              The data received, exactly ``num_bytes`` bytes;
              unless EOF is encountered before reading any data, in which
              case we return ``None``.

        Raises:
          ValueError: if EOF is encountered after reading at least one byte
              but before reading ``num_bytes`` bytes.
        """
        data = await self.receive(num_bytes)
        if data == b"":
            return None
        if len(data) != num_bytes:
            self._receive_pos -= len(data)
            raise ValueError(
                f"unclean EOF ({len(data)} bytes after boundary, "
                f"expected at least {num_bytes})"
            )
        return data

    async def receive_exactly(self, num_bytes: int) -> bytes:
        """Receive and return exactly ``num_bytes`` bytes, throwing an
        exception if EOF is encountered before then.

        Args:
          num_bytes (int): The number of bytes to return. Must be
              greater than zero.

        Returns:
          bytes: The data received, exactly ``num_bytes`` bytes.

        Raises:
          ValueError: if EOF is encountered before reading ``num_bytes`` bytes.
        """
        data = await self.receive(num_bytes)
        if len(data) != num_bytes:
            self._receive_pos -= len(data)
            raise ValueError(f"unclean EOF (read only {len(data)}/{num_bytes} bytes)")
        return data

    def unget(self, data: bytes) -> None:
        """Put the bytes in ``data`` back into the buffer, so they will be the
        next thing received by a call to one of the receive methods.
        """
        new_receive_pos = max(0, self._receive_pos - len(data))
        self._buffer[new_receive_pos : self._receive_pos] = data
        self._receive_pos = new_receive_pos


class TextReceiveStream(trio.abc.AsyncResource):
    r"""Wraps a :class:`~trio.abc.ReceiveStream` with buffering and decoding
    capabilities for receiving line-oriented text.

    See :class:`io.TextIOWrapper` for more documentation on the ``encoding``,
    ``errors``, and ``newline`` arguments.

    Args:
      transport_stream (~trio.abc.ReceiveStream): The stream to receive
          data on.
      encoding (str): The encoding with which to decode received data.
          If none is specified, we use the value returned by
          :func:`locale.getpreferredencoding`.
      errors (str): Controls how to respond to decoding errors; common
          values include ``"strict"`` (throw an exception), ``"ignore"``
          (drop the bad character), or ``"replace"`` (replace the bad
          character with a replacement marker). The default of ``None``
          is equivalent to ``"strict"``.
      newline (str): Controls how line endings are handled. Use
          ``None`` to convert any newline format to ``"\n"``,
          ``""`` to accept any newline format and pass it through unchanged,
          or ``"\r"``, ``"\n"``, or ``"\r\n"`` to only accept that
          sequence as a newline.
      chunk_size (int): The number of bytes to request in each call to the
          underlying transport stream's
          :meth:`~trio.abc.ReceiveStream.receive_some` method.

    """

    transport_stream: trio.abc.ReceiveStream
    chunk_size: int

    # Either _decoder is the same as _underlying_decoder, or _decoder
    # is an IncrementalNewlineDecoder. We need to remember both
    # because IncrementalNewlineDecoder doesn't have a .errors
    # attribute.
    _decoder: codecs.IncrementalDecoder
    _underlying_decoder: codecs.IncrementalDecoder

    def __init__(
        self,
        transport_stream: trio.abc.ReceiveStream,
        encoding: Optional[str] = None,
        *,
        errors: Optional[str] = None,
        newline: Optional[str] = "",
        chunk_size: int = 8192,
    ):
        if encoding is None:
            import locale

            encoding = locale.getpreferredencoding(False)

        self.transport_stream = transport_stream
        self.chunk_size = chunk_size
        self._encoding = encoding

        # The newline parameter is a newline sequence, or "" to accept
        # any of \r \n \r\n, or None to convert all to \n. self._newline
        # is the sequence we'll look for, or "" for any of \r \n \r\n.

        self._newline = newline if newline is not None else "\n"

        def make_decoder(
            encoding: str, errors: Optional[str], universal: bool, translate: bool
        ) -> codecs.IncrementalDecoder:
            info = codecs.lookup(encoding)
            decoder = info.incrementaldecoder(errors)  # type: ignore
            self._underlying_decoder = decoder
            if universal:
                return IncrementalNewlineDecoder(decoder, translate)
            return decoder

        self._decoder = make_decoder(
            encoding, errors, newline is None or newline == "", newline is None
        )

        # Data that has been received but not yet passed through
        # self._decoder. We store it as a member variable to permit recovery
        # from decode errors; unless one of those occurs, it will be None
        # at every checkpoint.
        self._raw_chunk: Optional[bytes] = None

        # self._chunk[self._chunk_pos:] is the data that has been
        # passed through self._decoder but not yet returned from
        # receive_line().
        self._chunk = ""
        self._chunk_pos = 0

    @property
    def encoding(self) -> str:
        return self._encoding

    @property
    def errors(self) -> Optional[str]:
        return self._underlying_decoder.errors

    @errors.setter
    def errors(self, value: Optional[str]) -> None:
        self._underlying_decoder.errors = value  # type: ignore

    @property
    def newlines(self) -> Union[str, Tuple[str, ...], None]:
        r"""The newline sequences that have actually been observed in the input.

        If no newline sequences have been observed, *or* if you specified
        a particular ``newline`` type when constructing this stream,
        this attribute is ``None``. Otherwise, it is a single string
        or a tuple of strings drawn from the set ``{"\r", "\n", "\r\n"}``.
        """
        try:
            return self._decoder.newlines  # type: ignore
        except AttributeError:
            return None

    async def aclose(self) -> None:
        """Discard all buffered data and close the underlying stream."""
        self._raw_chunk = None
        self._chunk = ""
        self._chunk_pos = 0
        await self.transport_stream.aclose()

    async def __aiter__(self) -> AsyncIterator[str]:
        """Iterate over the lines in this stream."""
        while True:
            line = await self.receive_line()
            if line == "":
                return
            yield line

    async def receive_line(self, max_chars: int = -1) -> str:
        """Receive and decode data on this stream until ``max_chars`` have
        been received or a newline or end-of-file is encountered. The
        meaning of "newline" depends on the ``newline`` argument
        passed at construction time.

        Args:
          max_chars (int): The maximum number of characters to return if
              no newline sequence is received. If negative, read until
              newline or EOF.

        Returns:
          str: The line received. It always ends with a newline unless
          we reached ``max_chars`` or EOF. If there is no data left to
          return before EOF, returns an empty string (``""``).

        Raises:
          UnicodeDecodeError: if the received data can't be decoded
          Anything else: that was raised by the underlying transport stream's
            :meth:`~trio.abc.ReceiveStream.receive_some` method.

        """

        await trio.lowlevel.checkpoint_if_cancelled()

        got_more = False
        line_end_pos = None
        while True:
            max_pos = len(self._chunk)
            if max_chars > 0:
                max_pos = min(max_pos, self._chunk_pos + max_chars)

            if self._newline == "":
                # Universal newlines without translation: search for any of \r,
                # \n, \r\n. Use of IncrementalNewlineDecoder ensures we never
                # split a \r\n sequence across two decoder outputs.
                crpos = self._chunk.find("\r", self._chunk_pos, max_pos)
                lfpos = self._chunk.find("\n", self._chunk_pos, max_pos)
                if crpos != -1 or lfpos != -1:
                    # Found a newline
                    if crpos != -1 and (lfpos == -1 or crpos < lfpos):
                        # CR exists and comes before LF.  LF may or
                        # may not exist.  If the first LF is one
                        # position after the first CR, we have a CRLF
                        # and must end the line after the entire CRLF
                        # sequence. Otherwise, end after the CR.
                        line_end_pos = crpos + 1 + (lfpos == crpos + 1)
                    else:
                        # CR either does not exist or comes after LF,
                        # so this line is delimited by LF.
                        line_end_pos = lfpos + 1
                    break
            else:
                # Just need to end on occurrences of self._newline.
                # (If we're using universal newlines with translation, we
                # set it to "\n" in the constructor.)
                nlpos = self._chunk.find(self._newline, self._chunk_pos, max_pos)
                if nlpos != -1:
                    line_end_pos = nlpos + len(self._newline)
                    break

            # If we found a newline in self._chunk, we broke out of the
            # loop above. Getting here means we either need more data or
            # hit our max_chars limit and must return without the newline.
            if max_pos == self._chunk_pos + max_chars:
                # Hit limit, return what we've got.
                line_end_pos = max_pos
                break

            # Need to pull down more raw data to decode
            if self._raw_chunk is None:
                self._raw_chunk = await self.transport_stream.receive_some(
                    self.chunk_size
                )
                got_more = True

            if self._raw_chunk == b"":
                # EOF on underlying stream. Pull out whatever the decoder
                # has left for us; if that's nothing, return EOF ourselves.
                chunk = self._decoder.decode(self._raw_chunk, final=True)
                if not chunk:
                    line_end_pos = len(self._chunk)
                    break
            else:
                chunk = self._decoder.decode(self._raw_chunk)

            # We need to reallocate self._chunk in order to append the new
            # stuff, so we'll throw away already-consumed output while we're
            # at it. We don't do this at every call to receive_line() because
            # it would result in quadratic-time performance with short lines.
            # (We still get quadratic-time performance with arbitrarily long
            # lines, but we'll not worry about that for now.)
            self._chunk = self._chunk[self._chunk_pos :] + chunk
            self._chunk_pos = 0

            # We've incorporated _raw_chunk into _chunk, so null it out.
            # If decoding failed we would leave _raw_chunk non-null and
            # try again to decode it on a future call, maybe with a different
            # errors parameter.
            self._raw_chunk = None

        # We break out of the loop when we find the point we want to
        # chop at.  All that's left is to return it to the caller.

        if not got_more:
            # If we never called receive_some(), we only did half a checkpoint,
            # and need to do the other half before returning.
            await trio.lowlevel.cancel_shielded_checkpoint()

        ret = self._chunk[self._chunk_pos : line_end_pos]
        self._chunk_pos = line_end_pos

        # If we're consuming the whole buffer, compact it now since
        # that's basically free. Otherwise wait until we next pull
        # down a chunk, so we don't have too poor performance when
        # receiving lots of short lines.
        if self._chunk_pos == len(self._chunk):
            self._chunk = ""
            self._chunk_pos = 0

        return ret



from setuptools import setup, find_packages

exec(open("tricycle/_version.py", encoding="utf-8").read())

LONG_DESC = open("README.rst", encoding="utf-8").read()

setup(
    name="tricycle",
    version=__version__,
    description="Experimental extensions for Trio, the friendly async I/O library",
    url="https://github.com/oremanj/tricycle",
    long_description=LONG_DESC,
    author="Joshua Oreman",
    author_email="oremanj@gmail.com",
    license="MIT -or- Apache License 2.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["trio >= 0.15.0", "trio-typing >= 0.5.0"],
    keywords=["async", "trio"],
    python_requires=">=3.6",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Framework :: Trio",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
    ],
)



import attr
import trio
import weakref
from typing import Iterator, MutableSet, Optional


@attr.s(eq=False, repr=False)
class MultiCancelScope:
    r"""Manages a dynamic set of :class:`trio.CancelScope`\s that can be
    shielded and cancelled as a unit.

    New cancel scopes are added to the managed set using
    :meth:`open_child`, which returns the child scope so you can enter
    it with a ``with`` statement. Calls to :meth:`cancel` and changes
    to :attr:`shield` apply to all existing children and set the
    initial state for future children. Each child scope has its own
    :attr:`~trio.CancelScope.deadline` and :attr:`~trio.CancelScope.shield`
    attributes; changes to these do not modify the parent.

    There is no :attr:`~trio.CancelScope.cancelled_caught` attribute
    on :class:`MultiCancelScope` because it would be ambiguous; some
    of the child scopes might exit via a :exc:`trio.Cancelled`
    exception and others not. Look at the child :class:`trio.CancelScope`
    if you want to see whether it was cancelled or not.
    """

    _child_scopes: MutableSet[trio.CancelScope] = attr.ib(
        factory=weakref.WeakSet, init=False
    )
    _shield: bool = attr.ib(default=False, kw_only=True)
    _cancel_called: bool = attr.ib(default=False, kw_only=True)

    def __repr__(self) -> str:
        descr = ["MultiCancelScope"]
        if self._shield:
            descr.append(" shielded")
        if self._cancel_called:
            descr.append(" cancelled")
        return f"<{''.join(descr)}: {list(self._child_scopes)}>"

    @property
    def cancel_called(self) -> bool:
        """Returns true if :meth:`cancel` has been called."""
        return self._cancel_called

    @property
    def shield(self) -> bool:
        """The overall shielding state for this :class:`MultiCancelScope`.

        Setting this attribute sets the :attr:`~trio.CancelScope.shield`
        attribute of all children, as well as the default initial shielding
        for future children. Individual children may modify their
        shield state to be different from the parent value, but further
        changes to the parent :attr:`MultiCancelScope.shield` will override
        their local choice.
        """
        return self._shield

    @shield.setter
    def shield(self, new_value: bool) -> None:
        self._shield = new_value
        for scope in self._child_scopes:
            scope.shield = new_value

    def cancel(self) -> None:
        """Cancel all child cancel scopes.

        Additional children created after a call to :meth:`cancel` will
        start out in the cancelled state.
        """
        if not self._cancel_called:
            for scope in self._child_scopes:
                scope.cancel()
            self._cancel_called = True

    def open_child(self, *, shield: Optional[bool] = None) -> trio.CancelScope:
        """Return a new child cancel scope.

        The child will start out cancelled if the parent
        :meth:`cancel` method has been called. Its initial shield state
        is given by the ``shield`` argument, or by the parent's
        :attr:`shield` attribute if the ``shield`` argument is not specified.
        """
        if shield is None:
            shield = self._shield
        new_scope = trio.CancelScope(shield=shield)
        if self._cancel_called:
            new_scope.cancel()
        self._child_scopes.add(new_scope)
        return new_scope
